#include <stdio.h>
#include <stdlib.h>

#include "common.h"
#include "macros.h"
#include "failures.h"
#include "as_tree.h"
#include "symbol.h"
#include "quad.h"
#include "machine_generic.h"

/*
 * This architecture implementation was created only for debug purposes.
 * We just print quads as they are.
 */

extern code_print_out_t code_print_out;

static void print_arg_quad_const(struct quad_const *_const)
{
	return_if_fail(_const != NULL);

	switch (QUAD_ARG(_const)->data_type) {
	case DATA_TYPE_INTEGER:
		code_print_out("%d", _const->integer);
		break;
	case DATA_TYPE_REAL:
		code_print_out("%f", _const->real);
		break;
	case DATA_TYPE_CHAR:
		code_print_out("\'%c\'", _const->ch);
		break;
	case DATA_TYPE_STRING:
		code_print_out("\"%s\"", _const->str);
		break;
	default:
		break;
	}
}

static void print_arg_quad(struct quad *arg)
{
	return_if_fail(arg != NULL);
	return_if_fail(QUAD_ARG(arg)->data_type != DATA_TYPE_UNKNOWN);

	switch (QUAD(arg)->quad_type) {
	case QUAD_CONST:
		print_arg_quad_const(QUAD_CONST(arg));
		break;
	case QUAD_ID:
		code_print_out("%s", QUAD_ID(arg)->id);
		break;
	case QUAD_REFER:
		code_print_out("%s", QUAD_REFER(arg)->id);
		break;
	default:
		xerror(1, "unknown argument type");
	}
}

static void internal_label(struct quad *q, struct quads_ctx *ctx)
{
	struct quad_label *label_quad;

	return_if_fail(q != NULL);

	label_quad = (struct quad_label *)q;

	code_print_out(":L%d\n", label_quad->ln);
}

static void internal_goto(struct quad *q, struct quads_ctx *ctx)
{
	struct quad_goto *goto_quad;

	return_if_fail(q != NULL);

	goto_quad = (struct quad_goto *)q;

	code_print_out("goto L%d\n", goto_quad->ln);
}

static void internal_if(struct quad *q, struct quads_ctx *ctx)
{
	struct quad_if *if_quad;
	char *op;

	return_if_fail(q != NULL);

	if_quad = (struct quad_if *)q;

	switch (if_quad->opcode) {
	case OPCODE_GE:
		op = ">=";
		break;
	case OPCODE_GT:
		op = ">";
		break;
	case OPCODE_LE:
		op = "<=";
		break;
	case OPCODE_LT:
		op = "<";
		break;
	case OPCODE_NE:
		op = "!=";
		break;
	case OPCODE_EQ:
		op = "==";
		break;
	default:
		break;
	}

	code_print_out("if ");
	print_arg_quad(if_quad->left);
	code_print_out("%s", op);
	print_arg_quad(if_quad->right);
	code_print_out(" goto L%d\n", if_quad->ln);
}

static void internal_call(struct quad *q, struct quads_ctx *ctx)
{
	struct quad_call *call_quad;

	return_if_fail(q != NULL);

	call_quad = (struct quad_call *)q;

	code_print_out("call ");
	if (call_quad->ret_val != NULL) {
		print_arg_quad(call_quad->ret_val);
		code_print_out("=");
	}

	code_print_out("%s, %d\n", call_quad->name, call_quad->nargs);
}

static void internal_param(struct quad *q, struct quads_ctx *ctx)
{
	struct quad_param *param_quad;

	return_if_fail(q != NULL);

	param_quad = (struct quad_param *)q;

	code_print_out("param ");
	print_arg_quad(param_quad->param);
	code_print_out("\n");
}

static void internal_op(struct quad *q, struct quads_ctx *ctx)
{
	struct quad_op *op_quad;
	char op;

	return_if_fail(q != NULL);

	op_quad = (struct quad_op *)q;

	switch (op_quad->opcode) {
	case OPCODE_ADD:
		op = '+';
		break;
	case OPCODE_SUB:
		op = '-';
		break;
	case OPCODE_MUL:
		op = '*';
		break;
	case OPCODE_DIV:
		op = '/';
		break;
	case OPCODE_MOD:
		op = '%';
		break;
	default:
		break;
	}

	print_arg_quad(op_quad->result);
	code_print_out("=");
	print_arg_quad(op_quad->left);
	code_print_out("%c", op);
	print_arg_quad(op_quad->right);
	code_print_out("\n");
}

static void internal_assign(struct quad *q, struct quads_ctx *ctx)
{
	struct quad_assign *assign_quad;

	return_if_fail(q != NULL);

	assign_quad = (struct quad_assign *)q;

	print_arg_quad(assign_quad->lvalue);
	code_print_out("=");
	print_arg_quad(assign_quad->rvalue);
	code_print_out("\n");
}

static void internal_store_access(struct quad *q, struct quads_ctx *ctx)
{
	struct quad_store_access *store;

	return_if_fail(q != NULL);

	store = (struct quad_store_access *)q;

	code_print_out("%s[", store->name);
	print_arg_quad(store->idx);
	code_print_out("]=");
	print_arg_quad(store->rvalue);
	code_print_out("\n");
}

static void internal_load_access(struct quad *q, struct quads_ctx *ctx)
{
	struct quad_load_access *load;

	return_if_fail(q != NULL);

	load = (struct quad_load_access *)q;

	print_arg_quad(load->lvalue);
	code_print_out("=%s[", load->name);
	print_arg_quad(load->idx);
	code_print_out("]\n");
}

#ifdef _MEM_AREA
static void internal_load_mem_area(struct quad *q, struct quads_ctx *ctx)
{
	struct quad_load_mem_area *load;

	return_if_fail(q != NULL);

	load = (struct quad_load_mem_area *)q;

	print_arg_quad(load->lvalue);
	code_print_out("=$%c%d%c", load->area, load->offset, load->mem_data_type);

	if (load->idx != NULL) {
		code_print_out("[");
		print_arg_quad(load->idx);
		code_print_out("]");
	}

	if (load->nbits != -1)
		code_print_out(".%d", load->nbits);
	code_print_out("\n");
}

static void internal_store_mem_area(struct quad *q, struct quads_ctx *ctx)
{
	struct quad_store_mem_area *store;

	return_if_fail(q != NULL);

	store = (struct quad_store_mem_area *)q;

	code_print_out("$%c%d%c", store->area, store->offset, store->mem_data_type);

	if (store->idx != NULL) {
		code_print_out("[");
		print_arg_quad(store->idx);
		code_print_out("]");
	}

	if (store->nbits != -1)
		code_print_out(".%d", store->nbits);

	code_print_out("=");
	print_arg_quad(store->rvalue);
	code_print_out("\n");
}
#endif

static void internal_func(struct quad *q, struct quads_ctx *ctx)
{
	struct quad_func *func_quad;

	return_if_fail(q != NULL);

	func_quad = (struct quad_func *)q;

	code_print_out("%s:\n", func_quad->name);
}

static void internal_scope(struct quad *q, struct quads_ctx *ctx)
{
	struct quad_scope *scope_quad;
	struct symbol_table *global;

	return_if_fail(q != NULL);

	scope_quad = (struct quad_scope *)q;

	global = symbol_table_get_global_table();

	if (scope_quad->table != global)
		code_print_out("<<<\n");
}

static void internal_scope_end(struct quad *q, struct quads_ctx *ctx)
{
	struct quad_scope_end *scope_end_quad;

	return_if_fail(q != NULL);

	scope_end_quad = (struct quad_scope_end*)q;

	code_print_out(">>>\n");
}

static void internal_func_end(struct quad *q, struct quads_ctx *ctx)
{
	struct quad_func_end *func_end_quad;

	return_if_fail(q != NULL);

	func_end_quad = (struct quad_func_end*)q;

	code_print_out("ret\n");
	code_print_out("=========\n");
}

static struct q_vfuncs v_internal = {
	.q_label = internal_label,
	.q_goto = internal_goto,
	.q_if = internal_if,
	.q_call= internal_call,
	.q_param = internal_param,
	.q_op = internal_op,
	.q_assign = internal_assign,
	.q_store_access = internal_store_access,
	.q_load_access = internal_load_access,
#ifdef _MEM_AREA
	.q_store_mem_area = internal_store_mem_area,
	.q_load_mem_area = internal_load_mem_area,
#endif
	.q_func = internal_func,
	.q_scope = internal_scope,
	.q_scope_end = internal_scope_end,
	.q_func_end = internal_func_end
};

/* Set quads handlers. */
void machine_internal_set_vfuncs()
{
	machine_set_vfuncs(&v_internal);
}
